# Generated by Django 3.1.2 on 2020-10-15 20:30

from django.db import migrations, models
import django.db.models.deletion
import dynamic_storages.db.fields
import uuid


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='TestStorageTarget',
            fields=[
                ('id', models.UUIDField(default=uuid.uuid4, help_text='UUID identifying this objects', primary_key=True, serialize=False)),
                ('name', models.CharField(db_index=True, help_text='Name of this object', max_length=150)),
                ('description', models.TextField(blank=True, help_text='Description of this object', null=True)),
                ('last_checked', models.DateTimeField(blank=True, editable=False, help_text='Timestamp identifying when this storage provider was last checked', null=True)),
                ('last_status', models.CharField(choices=[['v', 'Valid'], ['e', 'Error'], ['u', 'Unknown']], default='u', editable=False, help_text='Flag indiciating what the last status check result was', max_length=1)),
                ('status_detail', models.TextField(blank=True, editable=False, help_text='Status details from last status check', null=True)),
                ('provider', models.CharField(choices=[('libcloud', 'Apache LibCloud'), ('azure', 'Azure Blob Storage'), ('dropbox', 'Dropbox'), ('ftp', 'FTP'), ('gcloud', 'Google Cloud Storage'), ('s3boto3', 'S3/Boto3'), ('do', 'Digital Ocean (boto3)'), ('sftp', 'SFTP')], default='gcloud', help_text='Specific storage provider this target utilizes - generally an object storage solution of some sort', max_length=8)),
                ('config', dynamic_storages.db.fields.EncryptedJSONField(blank=True, default=dict, help_text='Key/value pairs to pass to the storage provider when initializing the storage backend', null=True)),
                ('created', models.DateTimeField(auto_now_add=True, db_index=True, help_text='Timestamp indicating when this object was created')),
                ('modified', models.DateTimeField(auto_now=True, db_index=True, help_text='Timestamp indicating when this object was last updated')),
            ],
            options={
                'verbose_name': 'Storage Target',
                'verbose_name_plural': 'Storage Targets',
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='TestImageStorageModel',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('image', dynamic_storages.db.fields.DynamicStorageImageField(upload_to='')),
                ('storage_target', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='tests.teststoragetarget')),
            ],
        ),
        migrations.CreateModel(
            name='TestFileStorageModel',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('file', dynamic_storages.db.fields.DynamicStorageFileField(upload_to='')),
                ('storage_target', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='tests.teststoragetarget')),
            ],
        ),
    ]



from django.apps import AppConfig


class DynamicStoragesConfig(AppConfig):
    name = "dynamic_storages"
    label = "Dynamic Storages"
    verbose_name = "Dynamic Storages"

    def ready(self):
        pass



from django.test import TestCase
from .factories import TestStorageTargetFactory, TestFileStorageModelFactory, TestImageStorageModelFactory


class TestStorageTargetFactoryTestCase(TestCase):
    def test_gcs(self):
        tstf = TestStorageTargetFactory(provider="gcloud")
        self.assertIsNotNone(tstf.pk)

    def test_aws(self):
        tstf = TestStorageTargetFactory(provider="s3boto3")
        self.assertIsNotNone(tstf.pk)

    def test_azure(self):
        tstf = TestStorageTargetFactory(provider="azure")
        self.assertIsNotNone(tstf.pk)


class TestFileStorageModelFactoryTestCase(TestCase):
    def test_gcs(self):
        tstf = TestStorageTargetFactory(provider="gcloud")
        tfsmf = TestFileStorageModelFactory(storage_target=tstf)
        self.assertIsNotNone(tfsmf.pk)
        self.assertIsNotNone(tfsmf.file.url)

    def test_aws(self):
        tstf = TestStorageTargetFactory(provider="s3boto3")
        tfsmf = TestFileStorageModelFactory(storage_target=tstf)
        self.assertIsNotNone(tfsmf.pk)
        self.assertIsNotNone(tfsmf.file.url)

    def test_azure(self):
        tstf = TestStorageTargetFactory(provider="azure")
        tfsmf = TestFileStorageModelFactory(storage_target=tstf)
        self.assertIsNotNone(tfsmf.pk)
        self.assertIsNotNone(tfsmf.file.url)

class TestImageStorageModelFactoryTestCase(TestCase):
    def test_gcs(self):
        tstf = TestStorageTargetFactory(provider="gcloud")
        tismf = TestImageStorageModelFactory(storage_target=tstf)
        self.assertIsNotNone(tismf.pk)
        self.assertIsNotNone(tismf.image.url)

    def test_aws(self):
        tstf = TestStorageTargetFactory(provider="s3boto3")
        tismf = TestImageStorageModelFactory(storage_target=tstf)
        self.assertIsNotNone(tismf.pk)
        self.assertIsNotNone(tismf.image.url)

    def test_azure(self):
        tstf = TestStorageTargetFactory(provider="azure")
        tismf = TestImageStorageModelFactory(storage_target=tstf)
        self.assertIsNotNone(tismf.pk)
        self.assertIsNotNone(tismf.image.url)



from .models import TestStorageTarget, TestFileStorageModel, TestImageStorageModel
import base64
import json
import os
import string
from datetime import timedelta
from dynamic_storages.models.mappings import LAST_STATUS_CHOICES, STORAGE_PROVIDER_MAP
import factory
import factory.django
import factory.fuzzy
from django.utils import timezone
from django.utils.text import slugify
from faker import Faker
from functools import partial
from django.core.files.base import ContentFile
import logging
import requests

log = logging.getLogger(__name__)

fake = Faker()


def get_config(key, **kwargs):
    log.debug("Getting config for {}".format(key))
    ret_val = {}
    if key.lower() == "alc":
        pass
    elif key.lower() in ["as", "azure"]:
        ret_val = {
            "account_key": kwargs.get(
                "account_key",
                os.getenv(
                    "TEST_AZURE_ACCOUNT_KEY",
                ),
            ),
            "account_name": kwargs.get(
                "account_name",
                os.getenv("TEST_AZURE_ACCOUNT_NAME"),
            ),
            "overwrite_files": bool(kwargs.get("overwrite_files", "True")),
            "azure_container": kwargs.get(
                "azure_container",
                os.getenv("TEST_AZURE_CONTAINER"),
            ),
        }
    elif key.lower() == "dropbox":
        ret_val = {
            "oauth2_access_token": kwargs.get("oauth2_access_token", os.getenv("TEST_DBS_OAUTH2_TOKEN")),
            "root_path": kwargs.get("root_path", os.getenv("TEST_DBS_ROOT_PATH")),
        }
    elif key.lower() == "ftp":
        ret_val = {
            "location": "ftp://{}:{}@{}:{}/".format(
                kwargs.get("username", os.getenv("TEST_FTP_USER_NAME")),
                kwargs.get("password", os.getenv("TEST_FTP_USER_PASS")),
                kwargs.get("host", os.getenv("TEST_FTP_HOST")),
                kwargs.get("port", os.getenv("TEST_FTP_PORT")),
            )
        }
    elif key.lower() == "gcloud":
        ret_val = {
            "bucket_name": kwargs.get(
                "bucket_name",
                os.getenv("TEST_GCS_BUCKET_NAME"),
            ),
            "credentials": json.loads(base64.b64decode(os.getenv("TEST_GCS_SERVICE_ACCOUNT_JSON", "{}"))),
        }
    elif key.lower() == "do":
        ret_val = {
            "access_key": kwargs.get("access_key", os.getenv("TEST_DO_ACCESS_KEY")),
            "secret_key": kwargs.get("secret_key", os.getenv("TEST_DO_SECRET_KEY")),
            "bucket_name": kwargs.get(
                "bucket_name",
                os.getenv("TEST_DO_BUCKET_NAME"),
            ),
            "region_name": kwargs.get("region_name", os.getenv("TEST_DO_REGION_NAME", "sfo2")),
            "default_acl": kwargs.get("default_acl", "private"),
        }
    elif key.lower() in ["s3b", "aws", "s3boto3"]:
        ret_val = {
            "access_key": kwargs.get("access_key", os.getenv("TEST_AWS_ACCESS_KEY")),
            "secret_key": kwargs.get("secret_key", os.getenv("TEST_AWS_SECRET_KEY")),
            "file_overwrite": kwargs.get("file_overwrite", True),
            "bucket_name": kwargs.get(
                "bucket_name",
                os.getenv("TEST_AWS_BUCKET_NAME", "introsepect-data-django-test"),
            ),
            "region_name": kwargs.get("region_name", os.getenv("TEST_AWS_REGION_NAME")),
            "default_acl": kwargs.get("default_acl", "private"),
        }
    elif key.lower() == "sftp":
        SSH_KEY_PATH = os.path.join(TEST_DATA_BASE_DIR, "sftp")
        with open(os.path.join(SSH_KEY_PATH, "sftp-test"), "r") as f:
            ssh_private_key = f.read()

        ret_val = {
            "host": kwargs.get("host", os.getenv("TEST_SFTP_HOST")),
            "params": {
                "port": int(kwargs.get("port", os.getenv("TEST_SFTP_PORT"))),
                "username": kwargs.get("username", os.getenv("TEST_SFTP_USER_NAME")),
                "password": kwargs.get("password", os.getenv("TEST_SFTP_USER_PASS")),
                "pkey": ssh_private_key,
            },
            "interactive": None,
            "file_mode": None,
            "dir_mode": None,
            "uid": None,
            "gid": None,
            "known_host_file": None,
            "root_path": None,
            "base_url": None,
        }
    log.debug("Config for {} is {}".format(key, ret_val))
    return ret_val


class TestStorageTargetFactory(factory.django.DjangoModelFactory):
    name = factory.LazyFunction(fake.catch_phrase)
    description = factory.LazyFunction(partial(fake.paragraph, 3))
    created = factory.LazyFunction(fake.past_datetime)
    provider = factory.fuzzy.FuzzyChoice([x for x in STORAGE_PROVIDER_MAP.keys()], getter=lambda c: c)
    config = factory.LazyAttribute(lambda o: get_config(o.provider))
    last_status = factory.fuzzy.FuzzyChoice([x for x in LAST_STATUS_CHOICES], getter=lambda c: c[0])

    class Meta:
        model = TestStorageTarget


def gen_file():
    return ContentFile(fake.binary(), name=fake.file_name())


class TestFileStorageModelFactory(factory.django.DjangoModelFactory):
    storage_target = factory.SubFactory(TestStorageTargetFactory)
    file = factory.django.FileField(from_func=gen_file)

    class Meta:
        model = TestFileStorageModel


def gen_image_file():
    return ContentFile(requests.get(fake.image_url()).content, name=fake.file_name(extension='jpeg'))


class TestImageStorageModelFactory(factory.django.DjangoModelFactory):
    storage_target = factory.SubFactory(TestStorageTargetFactory)
    image = factory.django.FileField(from_func=gen_image_file)

    class Meta:
        model = TestImageStorageModel



import logging
from django.db.models.fields.files import FieldFile, FileField, ImageField
from django.core.files.images import ImageFile

import json
from fernet_fields import EncryptedTextField
from django.core.serializers.json import DjangoJSONEncoder

log = logging.getLogger(__name__)

class FileFieldMixin(object):
    def pre_save(self, model_instance, add):
        if getattr(self, '_storage_callable', None):
            if callable(self._storage_callable):
                self.storage = self._storage_callable(model_instance)
        return super().pre_save(model_instance, add)


class EncryptedJSONField(EncryptedTextField):
    """JSON dictionary that is encrypted at rest as a text object - based off of the django-fernet-fields EncryptedTextField"""
    def get_prep_value(self, value):
        ret_val = "{}"
        if value:
            ret_val = json.dumps(value, separators=(",", ":"), cls=DjangoJSONEncoder)
        return ret_val

    def to_python(self, value):
        ret_val = {}
        if isinstance(value, dict):
            ret_val = value
        elif isinstance(value, str):
            ret_val = json.loads(value)
        return ret_val

class DynamicStorageFieldFile(FieldFile):
    """FieldFile implementation for dynamically setting the storage provider for the file at runtime based on a callable passed into the field definition"""
    def __init__(self, instance, field, name):
        if getattr(field, '_storage_callable', None):
            storage = field._storage_callable(instance)
            log.info('Setting storage for this field to {}'.format(storage))
            field.storage = storage
        super().__init__(instance, field, name)


class DynamicStorageFileField(FileFieldMixin, FileField):
    """File field implementation for dynamically setting the storage provider for the file to be stored/retrieved at runtime based on a callable passed into this field definition for the `storage` kwarg"""
    attr_class = DynamicStorageFieldFile


class DynamicStorageImageFieldFile(ImageFile, DynamicStorageFieldFile):
    def delete(self, save=True):
        # Clear the image dimensions cache
        if hasattr(self, '_dimensions_cache'):
            del self._dimensions_cache
        super().delete(save)

class DynamicStorageImageField(FileFieldMixin, ImageField):
    """Image field implementation for dynamically setting the storage provider for the image to be stored/retrieved at runtime based on a callable passed into this field definition for the `storage` kwarg"""
    attr_class = DynamicStorageImageFieldFile



from dynamic_storages.models import AbstractStorageTarget
from django.db import models
from dynamic_storages.db.fields import DynamicStorageFileField, DynamicStorageImageField
from django.core.files.storage import default_storage


class TestStorageTarget(AbstractStorageTarget):
    class Meta(AbstractStorageTarget.Meta):
        abstract = False


def get_storage(instance=None):
    if instance:
        return instance.storage_target.storage_backend
    return default_storage


class TestFileStorageModel(models.Model):
    file = DynamicStorageFileField(storage=get_storage)
    storage_target = models.ForeignKey("TestStorageTarget", on_delete=models.CASCADE)

class TestImageStorageModel(models.Model):
    image = DynamicStorageImageField(storage=get_storage)
    storage_target = models.ForeignKey('TestStorageTarget', on_delete=models.CASCADE)



from google.oauth2 import service_account

from ..db.fields import EncryptedJSONField
from django.db import models
from django.utils import timezone
from django.utils.translation import ugettext_lazy as _
from .mappings import LAST_STATUS_CHOICES, STORAGE_PROVIDER_MAP
import logging
from uuid import uuid4

log = logging.getLogger(__name__)


class AbstractStorageTarget(models.Model):
    """Abstract implementation of a storage target which includes access details for how to interact with the downstream object storage system"""

    id = models.UUIDField(default=uuid4, primary_key=True, help_text=_("UUID identifying this objects"))
    name = models.CharField(max_length=150, db_index=True, help_text=_("Name of this object"))
    description = models.TextField(blank=True, null=True, help_text=_("Description of this object"))

    last_checked = models.DateTimeField(
        null=True,
        blank=True,
        editable=False,
        help_text=_("Timestamp identifying when this storage provider was last checked"),
    )
    last_status = models.CharField(
        max_length=1,
        default="u",
        editable=False,
        choices=LAST_STATUS_CHOICES,
        help_text=_("Flag indiciating what the last status check result was"),
    )
    status_detail = models.TextField(
        blank=True,
        null=True,
        editable=False,
        help_text=_("Status details from last status check"),
    )
    provider = models.CharField(
        max_length=8,
        default="gcloud",
        choices=((k, v.get("name")) for k, v in STORAGE_PROVIDER_MAP.items()),
        help_text=_("Specific storage provider this target utilizes - generally an object storage solution of some sort"),
    )
    config = EncryptedJSONField(
        default=dict,
        blank=True,
        null=True,
        help_text=_("Key/value pairs to pass to the storage provider when initializing the storage backend"),
    )
    created = models.DateTimeField(
        auto_now_add=True,
        db_index=True,
        help_text=_("Timestamp indicating when this object was created"),
    )
    modified = models.DateTimeField(
        auto_now=True,
        db_index=True,
        help_text=_("Timestamp indicating when this object was last updated"),
    )
    as_of = models.DateTimeField(
        default=timezone.now,
        db_index=True,
        help_text=_("Timestamp that this storage provider should be used as of"),
    )

    @classmethod
    def as_of(cls, timestamp):
        return cls.objects.filter(as_of__lte=timestamp)

    def __self__(self):
        return "{} - {}".format(self.name, self.get_storage_provider_display())

    @property
    def storage_backend(self):
        """Property that should be used when assigning a DynamicStorageFileField's `storage` property at runtime"""
        create_class = STORAGE_PROVIDER_MAP[self.provider]["class"]
        create_kwargs = getattr(self, "config", {}).copy()
        if self.provider == "gcloud":
            log.debug("creds: {}".format(create_kwargs.get("credentials", {})))
            create_kwargs["credentials"] = service_account.Credentials.from_service_account_info(create_kwargs.pop("credentials", {}))
        if self.provider == "digitalocean":
            create_kwargs["addressing_style"] = "path"
            if not create_kwargs.get("region_name"):
                create_kwargs["region_name"] = "SFO1"
            create_kwargs["endpoint_url"] = "https://{}.digitaloceanspaces.com".format(create_kwargs.get("region_name", "SFO1"))
        elif self.provider == "s3boto3":
            create_kwargs["default_acl"] = create_kwargs.get("default_acl", "bucket-owner-full-control")
        return create_class(**create_kwargs)

    def _check_credentials(self):
        pass

    class Meta:
        abstract = True
        verbose_name = _("Storage Target")
        verbose_name_plural = _("Storage Targets")



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['dynamic_storages',
 'dynamic_storages.db',
 'dynamic_storages.migrations',
 'dynamic_storages.models',
 'dynamic_storages.tasks',
 'dynamic_storages.tests',
 'dynamic_storages.tests.migrations']

package_data = \
{'': ['*']}

install_requires = \
['Pillow>=8.0.0,<9.0.0',
 'django-fernet-fields>=0.6,<0.7',
 'django-storages[libcloud,sftp,dropbox,boto3,google,azure]>=1.10.1,<2.0.0',
 'django>=3.0.0,<4.0.0']

setup_kwargs = {
    'name': 'django-dynamic-storages',
    'version': '0.1.1',
    'description': 'A collection of file fields and associated components to allow for dynamic configuration of storage properties for file-based fields within Django models.',
    'long_description': None,
    'author': 'Patrick McClory',
    'author_email': 'patrick@mcclory.io',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/mcclory/django-dynamic-storages',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from storages.backends import (
    apache_libcloud,
    azure_storage,
    dropbox,
    ftp,
    gcloud,
    s3boto3,
    sftpstorage,
)

LAST_STATUS_CHOICES = [["v", "Valid"], ["e", "Error"], ["u", "Unknown"]]
STORAGE_PROVIDER_MAP = {
    "libcloud": {"class": apache_libcloud.LibCloudStorage, "name": "Apache LibCloud"},
    "azure": {"class": azure_storage.AzureStorage, "name": "Azure Blob Storage"},
    "dropbox": {"class": dropbox.DropBoxStorage, "name": "Dropbox"},
    "ftp": {"class": ftp.FTPStorage, "name": "FTP"},
    "gcloud": {"class": gcloud.GoogleCloudStorage, "name": "Google Cloud Storage"},
    "s3boto3": {"class": s3boto3.S3Boto3Storage, "name": "S3/Boto3"},
    "do": {"class": s3boto3.S3Boto3Storage, "name": "Digital Ocean (boto3)"},
    "sftp": {"class": sftpstorage.SFTPStorage, "name": "SFTP"},
}
